@@ -13,26 +13,13 @@ let copy =
 
 let error ~loc fmt = Location.raise_errorf ~loc (Stdlib.( ^^ ) "ppx_globalize: " fmt)
 
-(* Attribute for marking local parameters *)
-let local_loc = Loc.make ~loc:Location.none "ocaml.local"
-
-let local_attr =
-  Ast_builder.Default.attribute ~loc:Location.none ~name:local_loc ~payload:(PStr [])
-;;
-
-(* Checking for attribute on global fields *)
-let has_global_attr attributes =
-  List.exists
-    ~f:(fun attr ->
-      String.equal attr.attr_name.txt "ocaml.global"
-      || String.equal attr.attr_name.txt "extension.global")
-    attributes
-;;
-
 let is_global_field ld =
   match ld.pld_mutable with
   | Mutable -> true
-  | Immutable -> has_global_attr ld.pld_attributes
+  | Immutable ->
+    (match Ppxlib_jane.Ast_builder.Default.get_label_declaration_modality ld with
+     | Some Global, _ -> true
+     | None, _ -> false)
 ;;
 
 (* Check if types are really recursive ignoring global and mutable
@@ -47,9 +34,6 @@ class is_recursive rec_flag decls =
 
 let really_recursive rec_flag decls = (new is_recursive rec_flag decls)#go ()
 
-(* Type with local attribute attached *)
-let ptyp_local typ = { typ with ptyp_attributes = local_attr :: typ.ptyp_attributes }
-
 (* The name of the globalize function for a given type name as a string *)
 let globalize_name type_name =
   if String.equal type_name "t" then "globalize" else "globalize_" ^ type_name
@@ -117,20 +101,20 @@ end = struct
         decl.ptype_params
         ~init:{ vars; params }
         ~f:(fun (typ, _) { vars; params } ->
-          let vars, sym =
-            match typ.ptyp_desc with
-            | Ptyp_var name ->
-              let prefix = "_globalize_" ^ name in
-              let sym = gen_symbol ~prefix () in
-              let vars = Map.add_exn vars ~key:name ~data:(Globalize (evar sym)) in
-              vars, sym
-            | _ ->
-              let prefix = "_globalize_param" in
-              let sym = gen_symbol ~prefix () in
-              vars, sym
-          in
-          let params = sym :: params in
-          { vars; params })
+        let vars, sym =
+          match typ.ptyp_desc with
+          | Ptyp_var name ->
+            let prefix = "_globalize_" ^ name in
+            let sym = gen_symbol ~prefix () in
+            let vars = Map.add_exn vars ~key:name ~data:(Globalize (evar sym)) in
+            vars, sym
+          | _ ->
+            let prefix = "_globalize_param" in
+            let sym = gen_symbol ~prefix () in
+            vars, sym
+        in
+        let params = sym :: params in
+        { vars; params })
     in
     t, t.params
   ;;
@@ -183,20 +167,18 @@ end = struct
   ;;
 end
 
+let globalize_arrow ~loc ty = [%type: [%t ty] -> [%t ty]]
+
 (* Generate the type for a copier function for a given list of type
    parameters and type name
 *)
 let generate_typ builder params type_name =
   let open (val builder : Ast_builder.S) in
+  let globalize_arrow = globalize_arrow ~loc in
   let type_lid = Located.lident type_name in
-  let type_ =
-    ptyp_arrow
-      Nolabel
-      (ptyp_local (ptyp_constr type_lid params))
-      (ptyp_constr type_lid params)
-  in
-  List.fold_right params ~init:type_ ~f:(fun param acc ->
-    ptyp_arrow Nolabel (ptyp_arrow Nolabel (ptyp_local param) param) acc)
+  let type_ = ptyp_constr type_lid params in
+  List.fold_right params ~init:(globalize_arrow type_) ~f:(fun param acc ->
+    ptyp_arrow Nolabel (globalize_arrow param) acc)
 ;;
 
 (* Is an object field a polymorphic method? *)
@@ -213,42 +195,47 @@ let is_polymorphic_method field =
    needing to worry about the scope of type variables. *)
 let rec type_head builder typ =
   let open (val builder : Ast_builder.S) in
-  match typ.ptyp_desc with
-  | Ptyp_any | Ptyp_var _ | Ptyp_extension _ -> ptyp_any
-  | Ptyp_tuple args ->
-    let args = List.map ~f:(fun _ -> ptyp_any) args in
-    ptyp_tuple args
-  | Ptyp_constr (lid, []) -> ptyp_constr (Located.mk lid.txt) []
-  | Ptyp_constr (lid, _ :: _) -> ptyp_constr (Located.mk lid.txt) [ ptyp_any ]
-  | Ptyp_variant (fields, closed, labels) ->
-    let fields =
-      List.map fields ~f:(fun field ->
-        match field.prf_desc with
-        | Rtag (label, const, args) ->
-          rtag label const (List.map ~f:(fun _ -> ptyp_any) args)
-        | Rinherit typ -> rinherit (type_head builder typ))
-    in
-    ptyp_variant fields closed labels
-  | Ptyp_alias (typ, _) -> type_head builder typ
-  | Ptyp_arrow (lbl, _, _) -> ptyp_arrow lbl ptyp_any ptyp_any
-  | Ptyp_package (mty, constrs) ->
-    let constrs = List.map ~f:(fun (lid, _) -> lid, ptyp_any) constrs in
-    ptyp_package (mty, constrs)
-  | Ptyp_object (fields, closed) ->
-    if List.exists fields ~f:is_polymorphic_method
-    then ptyp_any
-    else (
-      let fields =
-        List.map fields ~f:(fun field ->
-          match field.pof_desc with
-          | Otag (lbl, _) -> otag lbl ptyp_any
-          | Oinherit typ -> oinherit (type_head builder typ))
-      in
-      ptyp_object fields closed)
-  | Ptyp_class (lid, args) ->
-    let args = List.map ~f:(fun _ -> ptyp_any) args in
-    ptyp_class (Located.mk lid.txt) args
-  | Ptyp_poly _ -> assert false
+  match Ppxlib_jane.Jane_syntax.Core_type.of_ast typ with
+  | Some (Jtyp_tuple args, _) ->
+    let args = List.map ~f:(fun (lbl, _) -> lbl, ptyp_any) args in
+    Ppxlib_jane.Jane_syntax.Core_type.core_type_of ~loc ~attrs:[] (Jtyp_tuple args)
+  | Some (Jtyp_layout _, _) | None ->
+    (match typ.ptyp_desc with
+     | Ptyp_any | Ptyp_var _ | Ptyp_extension _ -> ptyp_any
+     | Ptyp_tuple args ->
+       let args = List.map ~f:(fun _ -> ptyp_any) args in
+       ptyp_tuple args
+     | Ptyp_constr (lid, []) -> ptyp_constr (Located.mk lid.txt) []
+     | Ptyp_constr (lid, _ :: _) -> ptyp_constr (Located.mk lid.txt) [ ptyp_any ]
+     | Ptyp_variant (fields, closed, labels) ->
+       let fields =
+         List.map fields ~f:(fun field ->
+           match field.prf_desc with
+           | Rtag (label, const, args) ->
+             rtag label const (List.map ~f:(fun _ -> ptyp_any) args)
+           | Rinherit typ -> rinherit (type_head builder typ))
+       in
+       ptyp_variant fields closed labels
+     | Ptyp_alias (typ, _) -> type_head builder typ
+     | Ptyp_arrow (lbl, _, _) -> ptyp_arrow lbl ptyp_any ptyp_any
+     | Ptyp_package (mty, constrs) ->
+       let constrs = List.map ~f:(fun (lid, _) -> lid, ptyp_any) constrs in
+       ptyp_package (mty, constrs)
+     | Ptyp_object (fields, closed) ->
+       if List.exists fields ~f:is_polymorphic_method
+       then ptyp_any
+       else (
+         let fields =
+           List.map fields ~f:(fun field ->
+             match field.pof_desc with
+             | Otag (lbl, _) -> otag lbl ptyp_any
+             | Oinherit typ -> oinherit (type_head builder typ))
+         in
+         ptyp_object fields closed)
+     | Ptyp_class (lid, args) ->
+       let args = List.map ~f:(fun _ -> ptyp_any) args in
+       ptyp_class (Located.mk lid.txt) args
+     | Ptyp_poly _ -> assert false)
 ;;
 
 let mode_crossing_attr_name = "globalized"
@@ -282,7 +269,7 @@ let globalized_mode_crossing exp typ loc =
           None
           (ppat_var { txt = "x"; loc })
           (pexp_ident { txt = Lident "x"; loc }))
-       (ptyp_arrow Nolabel (ptyp_local typ) (copy#core_type typ)))
+       [%type: [%t typ] -> [%t copy#core_type typ]])
     [ Nolabel, exp ]
 ;;
 
@@ -294,122 +281,139 @@ let rec generate_globalized_for_typ builder env exp name_opt typ =
   match Attribute.consume mode_crossing_attr_core_type typ with
   | Some (typ, ()) -> globalized_mode_crossing exp typ typ_loc
   | None ->
-    (match typ.ptyp_desc with
-     | Ptyp_var name ->
-       (match Env.lookup env name with
-        | Some (Globalize fn) -> eapply fn [ exp ]
-        | Some Universal ->
-          error
-            ~loc:typ.ptyp_loc
-            "Cannot generate globalize function for universal type variable '%s"
-            name
-        | None ->
-          error
-            ~loc:typ.ptyp_loc
-            "Cannot generate globalize function for unbound type variable '%s"
-            name)
-     | Ptyp_tuple args ->
+    (match Ppxlib_jane.Jane_syntax.Core_type.of_ast typ with
+     | Some (Jtyp_tuple args, _attrs) ->
        let tpat, texp = generate_globalized_for_tuple_args builder env args in
        pexp_let Nonrecursive [ value_binding ~pat:tpat ~expr:exp ] texp
-     | Ptyp_constr (lid, args) ->
-       let args =
-         List.map ~f:(generate_globalized_for_typ_as_function builder env None) args
-       in
-       let lid = globalize_lid lid.txt in
-       eapply (pexp_ident (Located.mk lid)) (args @ [ exp ])
-     | Ptyp_variant (fields, Closed, None) ->
-       let inherits, constants, nonconstants =
-         List.fold_right
-           fields
-           ~init:([], [], [])
-           ~f:(fun field (inherits, consts, nonconsts) ->
-             match field.prf_desc with
-             | Rtag (name, false, [ arg ]) -> inherits, consts, (name.txt, arg) :: nonconsts
-             | Rtag (name, true, []) -> inherits, name.txt :: consts, nonconsts
-             | Rtag (_, _, _) ->
-               error
-                 ~loc:typ.ptyp_loc
-                 "Cannot generate globalize function for partial variant type"
-             | Rinherit typ ->
-               (match typ.ptyp_desc with
-                | Ptyp_constr (lid, _) -> (lid.txt, typ) :: inherits, consts, nonconsts
-                | _ ->
-                  error
-                    ~loc:typ.ptyp_loc
-                    "Cannot generate globalize function for unnamed inherited variant \
-                     constructors"))
-       in
-       let inherit_cases =
-         List.map inherits ~f:(fun (lid, inher) ->
-           let v = gen_symbol ~prefix:"x" () in
-           let lid = Located.mk lid in
-           let lhs = ppat_alias (ppat_type lid) (Located.mk v) in
-           let typ =
-             match name_opt with
-             | None -> typ
-             | Some typ -> typ
-           in
-           let rhs =
-             pexp_coerce
-               (generate_globalized_for_typ builder env (evar v) None inher)
-               (Some (type_head builder inher))
-               (type_head builder typ)
-           in
-           case ~lhs ~rhs ~guard:None)
-       in
-       let constants_case =
-         match constants with
-         | [] -> None
-         | first :: rest ->
-           let v = gen_symbol ~prefix:"x" () in
-           let first_pat = ppat_variant first None in
-           let lhs =
-             ppat_alias
-               (List.fold ~init:first_pat rest ~f:(fun acc name ->
-                  ppat_or acc (ppat_variant name None)))
-               (Located.mk v)
-           in
-           let rhs = evar v in
-           Some (case ~lhs ~rhs ~guard:None)
-       in
-       let nonconstants_cases =
-         List.map nonconstants ~f:(fun (name, arg) ->
-           let v = gen_symbol ~prefix:"arg" () in
-           let lhs = ppat_variant name (Some (pvar v)) in
-           let arg = generate_globalized_for_typ builder env (evar v) None arg in
-           let rhs = pexp_variant name (Some arg) in
-           case ~lhs ~rhs ~guard:None)
-       in
-       let cases = inherit_cases @ Option.to_list constants_case @ nonconstants_cases in
-       pexp_match exp cases
-     | Ptyp_variant (_, Open, _) ->
-       error ~loc:typ.ptyp_loc "Cannot generate globalize function for open variant type"
-     | Ptyp_variant (_, Closed, Some _) ->
-       error
-         ~loc:typ.ptyp_loc
-         "Cannot generate globalize function for partial variant type"
-     | Ptyp_alias (typ, name) ->
-       (match Env.lookup env name with
-        | Some (Globalize fn) -> eapply fn [ exp ]
-        | Some Universal | None ->
-          generate_globalized_for_typ builder env exp name_opt typ)
-     | Ptyp_poly (names, typ) ->
-       let env = Env.enter_poly builder env names in
-       generate_globalized_for_typ builder env exp None typ
-     | Ptyp_any ->
-       error ~loc:typ.ptyp_loc "Cannot generate globalize function for unknown type"
-     | Ptyp_arrow (_, _, _) ->
-       error ~loc:typ.ptyp_loc "Cannot generate globalize function for function type"
-     | Ptyp_object (_, _) ->
-       error ~loc:typ.ptyp_loc "Cannot generate globalize function for object type"
-     | Ptyp_class (_, _) ->
-       error ~loc:typ.ptyp_loc "Cannot generate globalize function for class type"
-     | Ptyp_package _ ->
-       error
-         ~loc:typ.ptyp_loc
-         "Cannot generate globalize function for first-class module type"
-     | Ptyp_extension _ ->
-       error ~loc:typ.ptyp_loc "Cannot generate globalize function for unknown extension")
+     | Some (Jtyp_layout _, _) | None ->
+       (match typ.ptyp_desc with
+        | Ptyp_var name ->
+          (match Env.lookup env name with
+           | Some (Globalize fn) -> eapply fn [ exp ]
+           | Some Universal ->
+             error
+               ~loc:typ.ptyp_loc
+               "Cannot generate globalize function for universal type variable '%s"
+               name
+           | None ->
+             error
+               ~loc:typ.ptyp_loc
+               "Cannot generate globalize function for unbound type variable '%s"
+               name)
+        | Ptyp_tuple args ->
+          let tpat, texp =
+            generate_globalized_for_tuple_args
+              builder
+              env
+              (List.map ~f:(fun arg -> None, arg) args)
+          in
+          pexp_let Nonrecursive [ value_binding ~pat:tpat ~expr:exp ] texp
+        | Ptyp_constr (lid, args) ->
+          let args =
+            List.map ~f:(generate_globalized_for_typ_as_function builder env None) args
+          in
+          let lid = globalize_lid lid.txt in
+          eapply (pexp_ident (Located.mk lid)) (args @ [ exp ])
+        | Ptyp_variant (fields, Closed, None) ->
+          let inherits, constants, nonconstants =
+            List.fold_right
+              fields
+              ~init:([], [], [])
+              ~f:(fun field (inherits, consts, nonconsts) ->
+              match field.prf_desc with
+              | Rtag (name, false, [ arg ]) ->
+                inherits, consts, (name.txt, arg) :: nonconsts
+              | Rtag (name, true, []) -> inherits, name.txt :: consts, nonconsts
+              | Rtag (_, _, _) ->
+                error
+                  ~loc:typ.ptyp_loc
+                  "Cannot generate globalize function for partial variant type"
+              | Rinherit typ ->
+                (match typ.ptyp_desc with
+                 | Ptyp_constr (lid, _) -> (lid.txt, typ) :: inherits, consts, nonconsts
+                 | _ ->
+                   error
+                     ~loc:typ.ptyp_loc
+                     "Cannot generate globalize function for unnamed inherited variant \
+                      constructors"))
+          in
+          let inherit_cases =
+            List.map inherits ~f:(fun (lid, inher) ->
+              let v = gen_symbol ~prefix:"x" () in
+              let lid = Located.mk lid in
+              let lhs = ppat_alias (ppat_type lid) (Located.mk v) in
+              let typ =
+                match name_opt with
+                | None -> typ
+                | Some typ -> typ
+              in
+              let rhs =
+                pexp_coerce
+                  (generate_globalized_for_typ builder env (evar v) None inher)
+                  (Some (type_head builder inher))
+                  (type_head builder typ)
+              in
+              case ~lhs ~rhs ~guard:None)
+          in
+          let constants_case =
+            match constants with
+            | [] -> None
+            | first :: rest ->
+              let v = gen_symbol ~prefix:"x" () in
+              let first_pat = ppat_variant first None in
+              let lhs =
+                ppat_alias
+                  (List.fold ~init:first_pat rest ~f:(fun acc name ->
+                     ppat_or acc (ppat_variant name None)))
+                  (Located.mk v)
+              in
+              let rhs = evar v in
+              Some (case ~lhs ~rhs ~guard:None)
+          in
+          let nonconstants_cases =
+            List.map nonconstants ~f:(fun (name, arg) ->
+              let v = gen_symbol ~prefix:"arg" () in
+              let lhs = ppat_variant name (Some (pvar v)) in
+              let arg = generate_globalized_for_typ builder env (evar v) None arg in
+              let rhs = pexp_variant name (Some arg) in
+              case ~lhs ~rhs ~guard:None)
+          in
+          let cases =
+            inherit_cases @ Option.to_list constants_case @ nonconstants_cases
+          in
+          pexp_match exp cases
+        | Ptyp_variant (_, Open, _) ->
+          error
+            ~loc:typ.ptyp_loc
+            "Cannot generate globalize function for open variant type"
+        | Ptyp_variant (_, Closed, Some _) ->
+          error
+            ~loc:typ.ptyp_loc
+            "Cannot generate globalize function for partial variant type"
+        | Ptyp_alias (typ, name) ->
+          (match Env.lookup env name with
+           | Some (Globalize fn) -> eapply fn [ exp ]
+           | Some Universal | None ->
+             generate_globalized_for_typ builder env exp name_opt typ)
+        | Ptyp_poly (names, typ) ->
+          let env = Env.enter_poly builder env names in
+          generate_globalized_for_typ builder env exp None typ
+        | Ptyp_any ->
+          error ~loc:typ.ptyp_loc "Cannot generate globalize function for unknown type"
+        | Ptyp_arrow (_, _, _) ->
+          error ~loc:typ.ptyp_loc "Cannot generate globalize function for function type"
+        | Ptyp_object (_, _) ->
+          error ~loc:typ.ptyp_loc "Cannot generate globalize function for object type"
+        | Ptyp_class (_, _) ->
+          error ~loc:typ.ptyp_loc "Cannot generate globalize function for class type"
+        | Ptyp_package _ ->
+          error
+            ~loc:typ.ptyp_loc
+            "Cannot generate globalize function for first-class module type"
+        | Ptyp_extension _ ->
+          error
+            ~loc:typ.ptyp_loc
+            "Cannot generate globalize function for unknown extension"))
 
 (* Generate code for a function to globalize values of type [type]. *)
 and generate_globalized_for_typ_as_function builder env name_opt typ =
@@ -428,29 +432,31 @@ and generate_globalized_for_tuple_args builder env args =
   let pats, exps =
     List.fold_right
       ~init:([], [])
-      ~f:(fun arg (pats, exps) ->
+      ~f:(fun (lbl, arg) (pats, exps) ->
         let vin = gen_symbol ~prefix:"arg" () in
         let pat = pvar vin in
         let local_exp = evar vin in
         let exp =
-          if has_global_attr arg.ptyp_attributes
-          then local_exp
-          else generate_globalized_for_typ builder env local_exp None arg
+          match Ppxlib_jane.Ast_builder.Default.get_tuple_field_modality arg with
+          | Some Global, _ -> local_exp
+          | None, _ -> generate_globalized_for_typ builder env local_exp None arg
         in
-        pat :: pats, exp :: exps)
+        (lbl, pat) :: pats, (lbl, exp) :: exps)
       args
   in
   let pat =
     match pats with
-    | [] -> assert false
-    | [ pat ] -> pat
-    | _ :: _ -> ppat_tuple pats
+    | [] | [ (Some _, _) ] -> assert false
+    | [ (None, pat) ] -> pat
+    | _ :: _ ->
+      Ppxlib_jane.Jane_syntax.Pattern.pat_of ~loc ~attrs:[] (Jpat_tuple (pats, Closed))
   in
   let exp =
     match exps with
-    | [] -> assert false
-    | [ exp ] -> exp
-    | _ :: _ -> pexp_tuple exps
+    | [] | [ (Some _, _) ] -> assert false
+    | [ (None, exp) ] -> exp
+    | _ :: _ ->
+      Ppxlib_jane.Jane_syntax.Expression.expr_of ~loc ~attrs:[] (Jexp_tuple exps)
   in
   pat, exp
 ;;
@@ -502,16 +508,26 @@ let generate_globalized_for_variant builder env exp cds =
       cds
       ~init:([], [])
       ~f:(fun (cd : constructor_declaration) (consts, nonconsts) ->
-        match cd.pcd_args with
-        | Pcstr_tuple [] ->
-          let name = cd.pcd_name.txt in
-          let consts = name :: consts in
-          consts, nonconsts
-        | (Pcstr_tuple _ | Pcstr_record _) as args ->
-          let name = cd.pcd_name.txt in
-          let env = Env.enter_constructor_declaration builder env cd in
-          let nonconsts = (name, args, env) :: nonconsts in
-          consts, nonconsts)
+      (* We differentiate between constant cases for GADTs vs normal variants
+           because currently, the type checker does not allow the use of as-pattern
+           to rename an or-pattern of GADTs when it does allow us to do so for normal
+           variants.
+
+           This is fixed in an upstream PR: https://github.com/ocaml/ocaml/pull/11799
+
+           When this is merged we can collapse the constants case back into a singular
+           branch. *)
+      match cd.pcd_res, cd.pcd_args with
+      | None, Pcstr_tuple [] ->
+        let name = cd.pcd_name.txt in
+        let consts = name :: consts in
+        consts, nonconsts
+      | None, ((Pcstr_tuple _ | Pcstr_record _) as args)
+      | Some _, ((Pcstr_tuple _ | Pcstr_record _) as args) ->
+        let name = cd.pcd_name.txt in
+        let env = Env.enter_constructor_declaration builder env cd in
+        let nonconsts = (name, args, env) :: nonconsts in
+        consts, nonconsts)
   in
   let constants_case =
     match constants with
@@ -534,12 +550,22 @@ let generate_globalized_for_variant builder env exp cds =
     List.map nonconstants ~f:(fun (name, args, env) ->
       let pat, exp =
         match args with
-        | Pcstr_tuple args -> generate_globalized_for_tuple_args builder env args
-        | Pcstr_record lds -> generate_globalized_for_record_args builder env lds
+        | Pcstr_tuple [] -> None, None
+        | Pcstr_tuple args ->
+          let pat, exp =
+            generate_globalized_for_tuple_args
+              builder
+              env
+              (List.map ~f:(fun arg -> None, arg) args)
+          in
+          Some pat, Some exp
+        | Pcstr_record lds ->
+          let pat, exp = generate_globalized_for_record_args builder env lds in
+          Some pat, Some exp
       in
       let lid = Located.mk (Lident name) in
-      let lhs = ppat_construct lid (Some pat) in
-      let rhs = pexp_construct lid (Some exp) in
+      let lhs = ppat_construct lid pat in
+      let rhs = pexp_construct lid exp in
       case ~lhs ~rhs ~guard:None)
   in
   let cases = Option.to_list constants_case @ nonconstants_cases in
